#	Masrv
#	Copyright (C) 2007 Michal Vaner (vorner@ucw.cz)
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	he Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import glob
import splitter
import selector
from sender import sender
import config
import socket

class client( sender ):
	def __init__( self, sock, address, images, manager ):
		sender.__init__( self )
		self.__socket = sock
		self.__splitter = splitter.splitter( self )
		self.__hasCoords = False
		self.__ackPending = 0
		self.__images = images
		self.__lines = []
		self.__manager = manager
		self.__address = address
		( name, morenames, moreaddresses ) = socket.gethostbyaddr( address[ 0 ] )
		self.__searchaddrs = [ name ] + morenames + map( lambda x: x.split( '.' )[ 0 ], [ name ] + morenames ) + [ address[ 0 ] ] + moreaddresses + [ "" ]
		self.__lastUpdate = 0
		self.__striprange = int( config.data.get( "clientstrip", 1000 ) )
		self.__clienttimeout = int( config.data.get( "clienttimeout", 60 ) ) * 1000
		self.__lastcontact = glob.timestamp()
		manager.insert( self )

	def sockSend( self, data ):
		try:
			return self.__socket.send( data )
			self.__lastcontact = glob.timestamp()
		except:
			if not glob.daemon:
				import traceback
				traceback.print_exc()
			glob.warn( "Lost client due to send exception" )
			self.__clean()
			return 0

	def fileno( self ):
		return self.__socket.fileno()

	def err( self ):
		pass

	def read( self ):
		try:
			msg = self.__socket.recv( 4096 )
			self.__lastcontact = glob.timestamp()
		except:
			if not glob.daemon:
				import traceback
				traceback.print_exc()
			self.__clean()
			glob.warn( "Client lost due to recv exception" )
		else:
			if len( msg ) == 0:
				glob.info( "Socket closed" )
				self.__clean()
			else:
				glob.info( "Incoming: " + msg )
				self.__splitter.insert( msg )
				self.__processLines()

	def reportLine( self, line ):
		self.__lines.append( line )

	def __sendConfig( self, name ):
		conf = ""
		for i in self.__searchaddrs:
			index = i + ":" + name
			glob.info( "Trying client config " + index )
			if index in config.clients:
				conf = config.clients[ index ]
				glob.info( "Found it" )
				break
		self.enque( "C " + name + "\n" + conf + "\n" );

	def __processLines( self ):
		try:
			while self.__lines:
				cnt = 1
				ln = self.__lines[ 0 ].strip()
				if not ln:
					self.__lines[ :1 ] = []
					continue
				cmd = ln.split( None, 1 )[ 0 ]
				if cmd == "coords":
					( self.__left, self.__top, self.__width, self.__height ) = map( int, ln.split()[ 1:5 ] )
					self.__hasCoords = True
					self.__lastUpdate = 0
					self.__striprect = ( self.__left - self.__striprange, self.__top - self.__striprange, self.__left + self.__width + self.__striprange, self.__top + self.__height + self.__striprange )
				elif cmd == "get_img":
					if len( self.__lines ) <= 1:
						break;#Not enough info, leave it for the next time
					name = ln.split( None, 1 )[ 1 ]
					( flipx, flipy ) = map( lambda x: x == "1", self.__lines[ 1 ].strip().split() )
					self.__ackPending += 1
					self.enque( "I " + name + "\n" )
					data = self.__images.data( name, flipx, flipy )
					( w, h ) = self.__images.size( name )
					self.enque( " ".join( [ str( w ), str( h ), str( len( data ) ), str( int( flipx ) ), str( int( flipy ) ) ] ) + "\n" )
					self.enque( data )
					cnt = 2
				elif cmd == "config":
					name = ln.split( None, 1 )[ 1 ]
					self.__sendConfig( name )
				elif cmd == ".":
					if not self.__ackPending:
						raise Exception( "Ack for no action" )
					self.__ackPending -= 1
				else:
					raise Exception( "Unknown command " + cmd )
				self.__lines[ :cnt ] = []
		except:
			if not glob.daemon:
				import traceback
				traceback.print_exc()
			self.__clean()
			glob.warn( "Lost client due to exception" )

	def __clean( self ):
			selector.remove( self )
			self.__manager.remove( self )
			self.__socket.close()

	def wantUpdate( self, timestamp, mod ):
		if self.__lastcontact + self.__clienttimeout < timestamp:
			self.__clean()
			glob.warn( "Lost client due to timeout" )
		if not self.__hasCoords: return False
		if mod:
			self.__lastUpdate = 0
		return ( ( self.__ackPending == 0 ) and ( timestamp - self.__lastUpdate >= 1000 ) )

	def __printMoves( self, moves ):
		first = True
		for ( x, y ) in moves:
			if first:#Move to be relative to the client
				first = False
				x -= self.__left
				y -= self.__top
			self.enque( str( x ) + " " + str( y ) + "\n" )

	def __containobj( self, obj ):
		if obj[ 0 ] == "l":
			( ( oleft, otop ), owidth, oheight ) = ( obj[ 6 ][ 0 ], obj[ 3 ], obj[ 4 ] )
		elif obj[ 0 ] == "i":
			( ( oleft, otop ), owidth, oheight ) = ( obj[ 8 ][ 0 ], obj[ 4 ], obj[ 5 ] )
		if owidth < 0:
			oleft += owidth
			owidth *= -1
		if oheight < 0:
			otop += oheight
			oheight *= -1
		( oright, obottom ) = ( oleft + owidth, otop + oheight )
		( left, top, right, bottom ) = self.__striprect
		return( ( - min( - left, - oleft ) <= min( right, oright ) ) and ( - min( - top, - otop ) <= min( bottom, obottom ) ) )

	def pushUpdate( self, state, timestamp ):
		self.__lastUpdate = timestamp
		self.enque( "U {\n" )
		for obj in state:
			if not self.__containobj( obj ):
				continue
			if obj[ 0 ] == "l":
				( type, id, color, xdelta, ydelta, width, moves ) = obj
				self.enque( "l\n" + ( "".join( map( lambda x: hex( x ).replace( "0x", "", 1 ).upper().rjust( 2, "0" ), color ) ) ) + " " + str( xdelta ) + " " + str( ydelta ) + " " + str( width ) + "\n" + str( hash( id ) ) + "\n" + str( len( moves ) ) + "\n" )
				self.__printMoves( moves )
			elif obj[ 0 ] == "i":
				( type, id, images, time, width, height, flipX, flipY, moves ) = obj
				self.enque( "i\n" + str( len( images ) ) + " " + str( time ) + " " + str( width ) + " " + str( height ) + "\n" + "\n".join( images ) + "\n" + str( int( flipX ) ) + " " + str( int( flipY ) ) + "\n" + str( hash( id ) ) + "\n" + str( len( moves ) ) + "\n" )
				self.__printMoves( moves )
		self.enque( "}\n" )
		self.__ackPending += 1

